from acados_template import AcadosOcp, AcadosOcpSolver, AcadosSimSolver
from pendulum_model import export_pendulum_ode_model, export_parametric_pendulum_ode_model
import numpy as np
import scipy.linalg
from utils import plot_pendulum


X0 = np.array([0.0, np.pi/6, 0.0, 0.0])
N_horizon = 50
T_horizon = 2.0
Fmax = 80

def create_ocp_solver_description() -> AcadosOcp:
    # create ocp object to formulate the OCP
    ocp = AcadosOcp()

    # set model
    model = export_pendulum_ode_model()

    # TASK: make the model parametric and make the mass of the cart M a parameter.
    # the ocp solver description needs to know initial parameter values (ocp.parameter_values), which can be changed later on.
    # model = export_parametric_pendulum_ode_model()

    ocp.model = model

    # set dimensions
    ocp.dims.N = N_horizon

    # set cost
    Q_mat = 2*np.diag([1e3, 1e3, 1e-2, 1e-2])
    R_mat = 2*np.diag([1e-1])

    # the 'EXTERNAL' cost type can be used to define general cost terms
    # NOTE: This leads to additional (exact) hessian contributions when using GAUSS_NEWTON hessian.
    ocp.cost.cost_type = 'EXTERNAL'
    ocp.cost.cost_type_e = 'EXTERNAL'
    ocp.model.cost_expr_ext_cost = model.x.T @ Q_mat @ model.x + model.u.T @ R_mat @ model.u
    ocp.model.cost_expr_ext_cost_e = model.x.T @ Q_mat @ model.x

    # set constraints
    ocp.constraints.lbu = np.array([-Fmax])
    ocp.constraints.ubu = np.array([+Fmax])
    ocp.constraints.idxbu = np.array([0])

    ocp.constraints.x0 = X0

    # set options
    ocp.solver_options.qp_solver = 'PARTIAL_CONDENSING_HPIPM' # FULL_CONDENSING_QPOASES
    # PARTIAL_CONDENSING_HPIPM, FULL_CONDENSING_QPOASES, FULL_CONDENSING_HPIPM,
    # PARTIAL_CONDENSING_QPDUNES, PARTIAL_CONDENSING_OSQP, FULL_CONDENSING_DAQP
    ocp.solver_options.hessian_approx = 'GAUSS_NEWTON' # 'GAUSS_NEWTON', 'EXACT'
    ocp.solver_options.integrator_type = 'IRK'
    # ocp.solver_options.print_level = 1
    ocp.solver_options.nlp_solver_type = 'SQP' # SQP_RTI, SQP
    ocp.solver_options.nlp_solver_max_iter = 400
    # ocp.solver_options.levenberg_marquardt = 1e-4

    # set prediction horizon
    ocp.solver_options.tf = T_horizon

    return ocp



def solve_single_ocp():
    ocp = create_ocp_solver_description()
    acados_ocp_solver = AcadosOcpSolver(ocp, json_file = 'acados_ocp_' + ocp.model.name + '.json')

    nx = ocp.model.x.size()[0]
    nu = ocp.model.u.size()[0]
    simX = np.ndarray((N_horizon+1, nx))
    simU = np.ndarray((N_horizon, nu))

    status = acados_ocp_solver.solve()
    acados_ocp_solver.print_statistics() # encapsulates: stat = acados_ocp_solver.get_stats("statistics")

    if status != 0:
        raise Exception(f'acados returned status {status}.')

    # get solution
    for i in range(N_horizon):
        simX[i,:] = acados_ocp_solver.get(i, "x")
        simU[i,:] = acados_ocp_solver.get(i, "u")
    simX[N_horizon,:] = acados_ocp_solver.get(N_horizon, "x")

    plot_pendulum(np.linspace(0, T_horizon, N_horizon+1), Fmax, simU, simX, latexify=False)



def closed_loop_simulation():
    ocp = create_ocp_solver_description()
    acados_ocp_solver = AcadosOcpSolver(ocp, json_file = 'acados_ocp_' + ocp.model.name + '.json')

    # create an integrator with the same settings as used in the OCP solver.
    acados_integrator = AcadosSimSolver(ocp, json_file = 'acados_ocp_' + ocp.model.name + '.json')

    # TASK: once you have a parametric model,
    # use a different parameter value in the integrator
    # try to change the weight of the cart by 20%

    # acados_integrator.set(...)

    Nsim = 100

    nx = ocp.model.x.size()[0]
    nu = ocp.model.u.size()[0]
    simX = np.ndarray((Nsim+1, nx))
    simU = np.ndarray((Nsim, nu))

    xcurrent = X0
    simX[0,:] = xcurrent

    # closed loop
    for i in range(Nsim):

        # set initial state constraint
        acados_ocp_solver.set(0, "lbx", xcurrent)
        acados_ocp_solver.set(0, "ubx", xcurrent)

        # initialize solver
        for stage in range(N_horizon+1):
            acados_ocp_solver.set(stage, 'x', xcurrent)
        # for stage in range(N_horizon):
        #     acados_ocp_solver.set(stage, 'u', np.array([0.0]))

        # solve ocp
        status = acados_ocp_solver.solve()

        if status not in [0, 2]:
            acados_ocp_solver.print_statistics()
            plot_pendulum(np.linspace(0, T_horizon/N_horizon*i, i+1), Fmax, simU[:i,:], simX[:i+1,:])
            raise Exception(f'acados acados_ocp_solver returned status {status} in closed loop instance {i} with {xcurrent=}')

        simU[i,:] = acados_ocp_solver.get(0, "u")

        # simulate system
        acados_integrator.set("x", xcurrent)
        acados_integrator.set("u", simU[i,:])

        status = acados_integrator.solve()
        if status != 0:
            raise Exception(f'acados integrator returned status {status} in closed loop instance {i}')

        # update state
        xcurrent = acados_integrator.get("x")
        simX[i+1,:] = xcurrent

    # plot results
    plot_pendulum(np.linspace(0, T_horizon/N_horizon*Nsim, Nsim+1), Fmax, simU, simX)


if __name__ == "__main__":
    solve_single_ocp()
    # closed_loop_simulation()
